@@ -1,3 +1,21 @@
+/*  Copyright 1986-1992 Emmet P. Gray.
+ *  Copyright 1996-1998,2001,2002,2009 Alain Knaff.
+ *  This file is part of mtools.
+ *
+ *  Mtools is free software: you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
+ *
+ *  Mtools is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with Mtools.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
 #include "sysincludes.h"
 #include "msdos.h"
 #include "fsP.h"
@@ -32,7 +50,7 @@ int fat_free(Stream_t *Dir, unsigned int fat)
 
 int fatFreeWithDir(Stream_t *Dir, struct directory *dir)
 {
-	unsigned int first;
+	uint32_t first;
 
 	if((!strncmp(dir->name,".      ",8) ||
 	    !strncmp(dir->name,"..     ",8)) &&
@@ -43,7 +61,7 @@ int fatFreeWithDir(Stream_t *Dir, struct directory *dir)
 
 	first = START(dir);
   	if(fat32RootCluster(Dir))
-		first |= STARTHI(dir) << 16;
+		first |= (uint32_t) STARTHI(dir) << 16;
 	return fat_free(Dir, first);
 }
 
@@ -51,4 +69,3 @@ int fatFreeWithDirentry(direntry_t *entry)
 {
 	return fatFreeWithDir(entry->Dir, &entry->dir);
 }
-    
