@@ -34,7 +34,13 @@ typedef struct File_t {
 			   mt_off_t *res);
 	uint32_t FileSize;
 
+	/* How many bytes do we project to need for this file
+	   (includes those already in FileSize) */
 	uint32_t preallocatedSize;
+
+	/* How many clusters we have asked the lower layer to reserve
+	   for us (only what we will need in the future, excluding already
+	   allocated clusters in FileSize) */
 	uint32_t preallocatedClusters;
 
 	/* Absolute position of first cluster of file */
@@ -94,7 +100,6 @@ static int recalcPreallocSize(File_t *This)
 	unsigned int clus_size;
 	uint32_t neededPrealloc;
 	Fs_t *Fs = This->Fs;
-	int r;
 
 #if 0
 	if(This->FileSize & 0xc0000000) {
@@ -106,16 +111,21 @@ static int recalcPreallocSize(File_t *This)
 	}
 #endif
 	clus_size = Fs->cluster_size * Fs->sector_size;
-
 	currentClusters = filebytesToClusters(This->FileSize, clus_size);
 	neededClusters = filebytesToClusters(This->preallocatedSize, clus_size);
 	if(neededClusters < currentClusters)
 		neededPrealloc = 0;
 	else
 		neededPrealloc = neededClusters - currentClusters;
-	r = fsPreallocateClusters(Fs, neededPrealloc - This->preallocatedClusters);
-	if(r)
-		return r;
+	if(neededPrealloc > This->preallocatedClusters) {
+		int r = fsPreallocateClusters(Fs, neededPrealloc-
+					      This->preallocatedClusters);
+		if(r)
+			return r;
+	} else {
+		fsReleasePreallocateClusters(Fs, This->preallocatedClusters -
+					     neededPrealloc);
+	}
 	This->preallocatedClusters = neededPrealloc;
 	return 0;
 }
@@ -148,7 +158,7 @@ static unsigned int _countBlocks(Fs_t *This, unsigned int block)
 	unsigned int rel, oldabs, oldrel;
 
 	blocks = 0;
-	
+
 	oldabs = oldrel = rel = 0;
 
 	while (block <= This->last_fat && block != 1 && block) {
@@ -290,7 +300,7 @@ static int normal_map(File_t *This, uint32_t where, uint32_t *len, int mode,
 	}
 
 	RelCluNr = where / clus_size;
-	
+
 	if (RelCluNr >= This->PreviousRelCluNr){
 		CurCluNr = This->PreviousRelCluNr;
 		AbsCluNr = This->PreviousAbsCluNr;
@@ -314,7 +324,7 @@ static int normal_map(File_t *This, uint32_t where, uint32_t *len, int mode,
 				AbsCluNr, NewCluNr);
 			exit(1);
 		}
-		if(CurCluNr == RelCluNr + NrClu)			
+		if(CurCluNr == RelCluNr + NrClu)
 			break;
 		if (NewCluNr > Fs->last_fat && mode == MT_WRITE){
 			/* if at end, and writing, extend it */
@@ -342,7 +352,7 @@ static int normal_map(File_t *This, uint32_t where, uint32_t *len, int mode,
 	}
 
 	maximize(*len, (1 + CurCluNr - RelCluNr) * clus_size - offset);
-	
+
 	end = where + *len;
 	if(batchmode &&
 	   mode == MT_WRITE &&
@@ -356,9 +366,9 @@ static int normal_map(File_t *This, uint32_t where, uint32_t *len, int mode,
 		exit(1);
 	}
 
-	*res = sectorsToBytes((Stream_t*)Fs,
+	*res = sectorsToBytes(Fs,
 			      (This->PreviousAbsCluNr-2) * Fs->cluster_size +
-			      Fs->clus_start) + offset;
+			      Fs->clus_start) + to_mt_off_t(offset);
 	return 1;
 }
 
@@ -368,7 +378,7 @@ static int root_map(File_t *This, uint32_t where, uint32_t *len,
 {
 	Fs_t *Fs = This->Fs;
 
-	if(Fs->dir_len * Fs->sector_size + 0u < (size_t) where) {
+	if(Fs->dir_len * Fs->sector_size < where) {
 		*len = 0;
 		errno = ENOSPC;
 		return -2;
@@ -377,11 +387,12 @@ static int root_map(File_t *This, uint32_t where, uint32_t *len,
 	maximize(*len, Fs->dir_len * Fs->sector_size - where);
         if (*len == 0)
             return 0;
-	
-	*res = sectorsToBytes((Stream_t*)Fs, Fs->dir_start) + where;
+
+	*res = sectorsToBytes(Fs, Fs->dir_start) +
+		to_mt_off_t(where);
 	return 1;
 }
-	
+
 
 static ssize_t read_file(Stream_t *Stream, char *buf, mt_off_t iwhere,
 			 size_t ilen)
@@ -391,9 +402,9 @@ static ssize_t read_file(Stream_t *Stream, char *buf, mt_off_t iwhere,
 	int err;
 	uint32_t where = truncMtOffTo32u(iwhere);
 	uint32_t len = truncSizeTo32u(ilen);
-	
+
 	Stream_t *Disk = This->Fs->Next;
-	
+
 	err = This->map(This, where, &len, MT_READ, &pos);
 	if(err <= 0)
 		return err;
@@ -410,13 +421,14 @@ static ssize_t write_file(Stream_t *Stream, char *buf,
 	uint32_t bytesWritten;
 	Stream_t *Disk = This->Fs->Next;
 	uint32_t where = truncMtOffTo32u(iwhere);
+	uint32_t maxLen = UINT32_MAX-where;
 	uint32_t len;
 	int err;
 
-	if(iwhere+(mt_off_t)ilen > UINT32_MAX) {
-		ilen=(size_t) (UINT32_MAX-iwhere);
-	}
-	len = truncSizeTo32u(ilen);
+	if(ilen > maxLen) {
+		len = maxLen;
+	} else
+		len = (uint32_t) ilen;
 	requestedLen = len;
 	err = This->map(This, where, &len, MT_WRITE, &pos);
 	if( err <= 0)
@@ -471,7 +483,7 @@ static __inline__ time_t conv_stamp(struct directory *dir)
 	{
 		struct timeval tv;
 		struct timezone tz;
-		
+
 		gettimeofday(&tv, &tz);
 		tzone = tz.tz_minuteswest * 60L;
 	}
@@ -503,7 +515,7 @@ static __inline__ time_t conv_stamp(struct directory *dir)
 }
 
 
-static int get_file_data(Stream_t *Stream, time_t *date, mt_size_t *size,
+static int get_file_data(Stream_t *Stream, time_t *date, mt_off_t *size,
 			 int *type, uint32_t *address)
 {
 	DeclareThis(File_t);
@@ -511,7 +523,7 @@ static int get_file_data(Stream_t *Stream, time_t *date, mt_size_t *size,
 	if(date)
 		*date = conv_stamp(& This->direntry.dir);
 	if(size)
-		*size = (mt_size_t) This->FileSize;
+		*size = to_mt_off_t(This->FileSize);
 	if(type)
 		*type = This->direntry.dir.attr & ATTR_DIR;
 	if(address)
@@ -524,7 +536,7 @@ static int free_file(Stream_t *Stream)
 {
 	DeclareThis(File_t);
 	Fs_t *Fs = This->Fs;
-	fsPreallocateClusters(Fs, -This->preallocatedClusters);
+	fsReleasePreallocateClusters(Fs, This->preallocatedClusters);
 	FREE(&This->direntry.Dir);
 	freeDirCache(Stream);
 	return hash_remove(filehash, (void *) Stream, This->hint);
@@ -611,7 +623,7 @@ static int comp(void *Stream, void *Stream2)
 static void init_hash(void)
 {
 	static int is_initialised=0;
-	
+
 	if(!is_initialised){
 		make_ht(func1, func2, comp, 20, &filehash);
 		is_initialised = 1;
